#ifndef IF97HEADER_H
#define IF97HEADER_H

#include <vector>
#include <cmath>
#include <iostream>
#include <iomanip>      // std::setprecision
#include <stdexcept>

enum I97parameters {IF97_DMASS, IF97_HMASS, IF97_T, IF97_P, IF97_SMASS, IF97_UMASS, IF97_CPMASS, IF97_CVMASS, IF97_W};
struct RegionIdealElement
{
    int J; ///< The first index
    double n; ///< The leading numerical constant
};
struct RegionResidualElement
{
    int I, ///< The first index
        J; ///< The second index
    double n; ///< The leading numerical constant
};

namespace IF97
{    
    double Tsat97(double p);
    class BaseRegion
    {
    public:
        BaseRegion(std::vector<RegionResidualElement> resid, std::vector<RegionIdealElement> ideal) : R(461.526){
            for (std::size_t i = 0; i < resid.size(); ++i){
                nr.push_back(resid[i].n);
                Ir.push_back(resid[i].I);
                Jr.push_back(resid[i].J);
            }
            for (std::size_t i = 0; i < ideal.size(); ++i){
                n0.push_back(ideal[i].n);
                J0.push_back(ideal[i].J);
            }
        }
        double rhomass(double T, double p){
            double PI = p/p_star;
            return p/(PI*R*T*(dgamma0_dPI(T,p) + dgammar_dPI(T,p)));
        }
        double hmass(double T, double p){
            double tau = T_star/T;
            return R*T*tau*(dgamma0_dTAU(T, p) + dgammar_dTAU(T, p));
        }
        double smass(double T, double p){
            double tau = T_star/T;
            return R*(tau*(dgamma0_dTAU(T, p) + dgammar_dTAU(T, p)) - (gammar(T, p) + gamma0(T, p)));
        }
        double umass(double T, double p){
            double tau = T_star/T, PI = p/p_star;
            return R*T*(tau*(dgamma0_dTAU(T, p) + dgammar_dTAU(T, p)) - PI*(dgamma0_dPI(T, p) + dgammar_dPI(T, p)));
        }
        double cpmass(double T, double p){
            double tau = T_star/T;
            return -R*tau*tau*(d2gammar_dTAU2(T, p) + d2gamma0_dTAU2(T, p));
        }
        double cvmass(double T, double p){
            double tau = T_star/T, PI = p/p_star;
            return cpmass(T,p)-R*pow(1 + PI*dgammar_dPI(T,p) - tau*PI*d2gammar_dPIdTAU(T,p),2)/(1-PI*PI*d2gammar_dPI2(T, p));
        }
        virtual double speed_sound(double T, double p){
            double tau = T_star/T, PI = p/p_star;
            double RHS = (1 + 2*PI*dgammar_dPI(T,p) + PI*PI*pow(dgammar_dPI(T,p),2))/((1-PI*PI*d2gammar_dPI2(T,p)) +pow(1 + PI*dgammar_dPI(T,p) - tau*PI*d2gammar_dPIdTAU(T,p), 2)/(tau*tau*(d2gamma0_dTAU2(T,p) + d2gammar_dTAU2(T,p))));
            return sqrt(R*T*RHS);
        }
        virtual double PIrterm(double) = 0;
        virtual double TAUrterm(double) = 0;
        virtual double TAU0term(double) = 0;
        double output(I97parameters key, double T, double p){
            switch(key){
            case IF97_T: return T;
            case IF97_P: return p;
            case IF97_DMASS: return rhomass(T, p);
            case IF97_HMASS: return hmass(T, p);
            case IF97_SMASS: return smass(T, p);
            case IF97_UMASS: return umass(T, p);
            case IF97_CPMASS: return cpmass(T, p);
            case IF97_CVMASS: return cvmass(T, p);
            case IF97_W: return speed_sound(T, p);
            }
            throw std::out_of_range("Unable to match input parameters");
        }

    protected:
        std::vector<int> Ir, Jr;
        std::vector<double> nr;
        std::vector<int> J0;
        std::vector<double> n0;
        double T_star, p_star;
        const double R;
   
        double gammar(double T, double p){
            double _PI = PIrterm(p), _TAU = TAUrterm(T), summer = 0;
            for (std::size_t i = 0; i < Jr.size(); ++i){
                summer += nr[i]*pow(_PI, Ir[i])*pow(_TAU, Jr[i]);
            }
            return summer;
        }
        double dgammar_dPI(double T, double p){
            double _PI = PIrterm(p), _TAU = TAUrterm(T), summer = 0;
            for (std::size_t i = 0; i < Jr.size(); ++i){
                summer += nr[i]*Ir[i]*pow(_PI, Ir[i]-1)*pow(_TAU, Jr[i]);
            }
            return summer;
        }
        double d2gammar_dPI2(double T, double p){
            double _PI = PIrterm(p), _TAU = TAUrterm(T), summer = 0;
            for (std::size_t i = 0; i < Jr.size(); ++i){
                summer += nr[i]*Ir[i]*(Ir[i]-1)*pow(_PI, Ir[i]-2)*pow(_TAU, Jr[i]);
            }
            return summer;
        }
        double dgammar_dTAU(double T, double p){
            double _PI = PIrterm(p), _TAU = TAUrterm(T), summer = 0;
            for (std::size_t i = 0; i < Jr.size(); ++i){
                summer += nr[i]*Jr[i]*pow(_PI, Ir[i])*pow(_TAU, Jr[i]-1);
            }
            return summer;
        }
        double d2gammar_dPIdTAU(double T, double p){
            double _PI = PIrterm(p), _TAU = TAUrterm(T), summer = 0;
            for (std::size_t i = 0; i < Jr.size(); ++i){
                summer += nr[i]*Jr[i]*Ir[i]*pow(_PI, Ir[i]-1)*pow(_TAU, Jr[i]-1);
            }
            return summer;
        }
        double d2gammar_dTAU2(double T, double p){
            double _PI = PIrterm(p), _TAU = TAUrterm(T), summer = 0;
            for (std::size_t i = 0; i < Jr.size(); ++i){
                summer += nr[i]*Jr[i]*(Jr[i]-1)*pow(_PI, Ir[i])*pow(_TAU, Jr[i]-2);
            }
            return summer;
        }
        double gamma0(double T, double p){
            if (J0.size() == 0){ return 0; } // Region 1 has no term
            double PI = p/p_star, _TAU = TAU0term(T);
            double summer = log(PI);
            for (std::size_t i = 0; i < n0.size(); ++i){
                summer += n0[i]*pow(_TAU, J0[i]);
            }
            return summer;
        }
        double dgamma0_dPI(double T, double p){
            if (J0.size() == 0){ return 0; } // Region 1 has no term
            double PI = p/p_star;
            return 1/PI;
        }
        double d2gamma0_dPI2(double T, double p){
            if (J0.size() == 0){ return 0; } // Region 1 has no term
            double PI = p/p_star;
            return -1/(PI*PI);
        }
        double dgamma0_dTAU(double T, double p){
            double _TAU = TAU0term(T), summer = 0;
            for (std::size_t i = 0; i < J0.size(); ++i){
                summer += n0[i]*J0[i]*pow(_TAU, J0[i]-1);
            }
            return summer;
        }
        double d2gamma0_dTAU2(double T, double p){
            double _TAU = TAU0term(T), summer = 0;
            for (std::size_t i = 0; i < J0.size(); ++i){
                summer += n0[i]*J0[i]*(J0[i]-1)*pow(_TAU, J0[i]-2);
            }
            return summer;
        }
    };

    /********************************************************************************/
    /**************************       Region #1       *******************************/
    /********************************************************************************/
    static RegionResidualElement Region1residdata[] = {
        // Note: the coefficients of n_i have been multiplied by -1**I_i such that all Gibbs terms are of the form (PI-7.1)**(I_i) rather than (7.1-PI)**(I_i)
        {0, -2, 0.14632971213167},
        {0, -1, -0.84548187169114},
        {0, 0, -3.756360367204},
        {0, 1, 3.3855169168385},
        {0, 2, -0.95791963387872},
        {0, 3, 0.15772038513228},
        {0, 4, -0.016616417199501},
        {0, 5, 0.00081214629983568},
        {1, -9, -0.00028319080123804},
        {1, -7, 0.00060706301565874},
        {1, -1, 0.018990068218419},
        {1, 0, 0.032529748770505},
        {1, 1, 0.021841717175414},
        {1, 3, 0.00005283835796993},
        {2, -3, -0.00047184321073267},
        {2, 0, -0.00030001780793026},
        {2, 1, 0.000047661393906987},
        {2, 3, -4.4141845330846E-06},
        {2, 17, -7.2694996297594E-16},
        {3, -4, 0.000031679644845054},
        {3, 0, 2.8270797985312E-06},
        {3, 6, 8.5205128120103E-10},
        {4, -5, -0.0000022425281908},
        {4, -2, -6.5171222895601E-07},
        {4, 10, -1.4341729937924E-13},
        {5, -8, 4.0516996860117E-07},
        {8, -11, -1.2734301741641E-09},
        {8, -6, -1.7424871230634E-10},
        {21, -29, 6.8762131295531E-19},
        {23, -31, -1.4478307828521E-20},
        {29, -38, -2.6335781662795E-23},
        {30, -39, -1.1947622640071E-23},
        {31, -40, -1.8228094581404E-24},
        {32, -41, -9.3537087292458E-26}
    };
    static std::vector<RegionResidualElement> reg1rdata(Region1residdata, Region1residdata + sizeof(Region1residdata)/sizeof(RegionResidualElement));
    static std::vector<RegionIdealElement> reg10data;
    class Region1 : public BaseRegion
    {
    public:
        Region1() : BaseRegion(reg1rdata, reg10data)  {
            T_star = 1386; p_star = 16.53e6; 
        };    
        double speed_sound(double T, double p){
            /// Evidently this formulation is special for some reason, and cannot be implemented using the base class formulation
            double tau = T_star/T;
            double RHS = pow(dgammar_dPI(T,p), 2)/(pow(dgammar_dPI(T,p)-tau*d2gammar_dPIdTAU(T,p), 2)/(tau*tau*d2gammar_dTAU2(T,p)) - d2gammar_dPI2(T, p));
            return sqrt(R*T*RHS);
        }
        double TAUrterm(double T){
            return T_star/T - 1.222;
        }
        double PIrterm(double p){
            return p/p_star - 7.1;
        }
        double TAU0term(double T){return 0.0;}
    };


    /********************************************************************************/
    /**************************       Region #2       *******************************/
    /********************************************************************************/
    static RegionResidualElement Region2residdata[] = {
        {1,0,-0.0017731742473213},
        {1,1,-0.017834862292358},
        {1,2,-0.045996013696365},
        {1,3,-0.057581259083432},
        {1,6,-0.05032527872793},
        {2,1,-0.000033032641670203},
        {2,2,-0.00018948987516315},
        {2,4,-0.0039392777243355},
        {2,7,-0.043797295650573},
        {2,36,-0.000026674547914087},
        {3,0,2.0481737692309E-08},
        {3,1,4.3870667284435E-07},
        {3,3,-0.00003227767723857},
        {3,6,-0.0015033924542148},
        {3,35,-0.040668253562649},
        {4,1,-7.8847309559367E-10},
        {4,2,1.2790717852285E-08},
        {4,3,4.8225372718507E-07},
        {5,7,2.2922076337661E-06},
        {6,3,-1.6714766451061E-11},
        {6,16,-0.0021171472321355},
        {6,35,-23.895741934104},
        {7,0,-5.905956432427E-18},
        {7,11,-1.2621808899101E-06},
        {7,25,-0.038946842435739},
        {8,8,1.1256211360459E-11},
        {8,36,-8.2311340897998},
        {9,13,1.9809712802088E-08},
        {10,4,1.0406965210174E-19},
        {10,10,-1.0234747095929E-13},
        {10,14,-1.0018179379511E-09},
        {16,29,-8.0882908646985E-11},
        {16,50,0.10693031879409},
        {18,57,-0.33662250574171},
        {20,20,8.9185845355421E-25},
        {20,35,3.0629316876232E-13},
        {20,48,-4.2002467698208E-06},
        {21,21,-5.9056029685639E-26},
        {22,53,3.7826947613457E-06},
        {23,39,-1.2768608934681E-15},
        {24,26,7.3087610595061E-29},
        {24,40,5.5414715350778E-17},
        {24,58,-9.436970724121E-07}
    };

    static RegionIdealElement Region2idealdata[] = {
    {0, -0.96927686500217e1},
    {1, 0.10086655968018e2},
    {-5, -0.56087911283020e-2 },
    {-4, 0.71452738081455e-1},
    {-3, -0.40710498223928},
    {-2, 0.14240819171444e1},
    {-1, -0.43839511319450e1},
    {2, -0.28408632460772},
    {3, 0.21268463753307e-1},
    };
    static std::vector<RegionResidualElement> reg2rdata(Region2residdata, Region2residdata + sizeof(Region2residdata)/sizeof(RegionResidualElement));
    static std::vector<RegionIdealElement> reg20data(Region2idealdata, Region2idealdata + sizeof(Region2idealdata)/sizeof(RegionIdealElement));
    class Region2 : public BaseRegion
    {
    public:
        Region2() : BaseRegion(reg2rdata, reg20data)  {
            T_star = 540; p_star = 1e6; 
        };
        double TAUrterm(double T){
            return T_star/T - 0.5;
        }
        double PIrterm(double p){
            return p/p_star;
        }
        double TAU0term(double T){return T_star/T;}
    };

    static double Region23data[] = {
    0.34805185628969e3,
    -0.11671859879975e1, 
    0.10192970039326e-2,
    0.57254459862746e3,
    0.13918839778870e2
    };

    static const std::vector<double> region23_n(Region23data, Region23data + sizeof(Region23data)/sizeof(double));

    inline double Region23_T(double T){
        const double p_star = 1e6, T_star = 1, theta = T/T_star;
        double PI = region23_n[0] + region23_n[1]*theta + region23_n[2]*theta*theta;
        return PI*p_star;
    }
    inline double Region23_p(double p){
        const double p_star = 1e6, T_star = 1, PI = p/p_star;
        double THETA = region23_n[4] + sqrt((PI - region23_n[5])/region23_n[3]);
        return THETA*T_star;
    }

    /********************************************************************************/
    /*********************       Region #3 (Backwards)      *************************/
    /********************************************************************************/

    namespace Region3Backwards{
        static RegionResidualElement Region3Adata[] = {
        {-12, 5, 0.110879558823853e-2},
        {-12, 10, 0.572616740810616e3},
        {-12, 12, -0.767051948380852e5},
        {-10, 5, -0.253321069529674e-1},
        {-10, 10, 0.628008049345689e4},
        {-10, 12, 0.234105654131876e6},
        {-8, 5, 0.216867826045856},
        {-8, 8, -0.156237904341963e3},
        {-8, 10, -0.269893956176613e5},
        {-6, 1, -0.180407100085505e-3},
        {-5, 1, 0.116732227668261e-2},
        {-5, 5, 0.266987040856040e2},
        {-5, 10, 0.282776617243286e5},
        {-4, 8, -0.242431520029523e4},
        {-3, 0, 0.435217323022733e-3},
        {-3, 1, -0.122494831387441e-1},
        {-3, 3, 0.179357604019989e1},
        {-3, 6, 0.442729521058314e2},
        {-2, 0, -0.593223489018342e-2},
        {-2, 2, 0.453186261685774},
        {-2, 3, 0.135825703129140e1},
        {-1, 0, 0.408748415856745e-1},
        {-1, 1, 0.474686397863312},
        {-1, 2, 0.118646814997915e1},
        {0, 0, 0.546987265727549},
        {0, 1, 0.195266770452643},
        {1, 0, -0.502268790869663e-1},
        {1, 2, -0.369645308193377},
        {2, 0, 0.633828037528420e-2},
        {2, 2, 0.797441793901017e-1},
        };

        static RegionResidualElement Region3Bdata[] = {
        {-12, 10, -0.827670470003621e-1},
        {-12, 12, 0.416887126010565e2},
        {-10, 8, 0.483651982197059e-1},
        {-10, 14, -0.291032084950276e5},
        {-8, 8, -0.111422582236948e3},
        {-6, 5, -0.202300083904014e-1},
        {-6, 6, 0.294002509338515e3},
        {-6, 8, 0.140244997609658e3},
        {-5, 5, -0.344384158811459e3},
        {-5, 8, 0.361182452612149e3},
        {-5, 10, -0.140699677420738e4},
        {-4, 2, -0.202023902676481e-2},
        {-4, 4, 0.171346792457471e3},
        {-4, 5, -0.425597804058632e1},
        {-3, 0, 0.691346085000334e-5},
        {-3, 1, 0.151140509678925e-2},
        {-3, 2, -0.416375290166236e-1},
        {-3, 3, -0.413754957011042e2},
        {-3, 5, -0.506673295721637e2},
        {-2, 0, -0.572212965569023e-3},
        {-2, 2, 0.608817368401785e1},
        {-2, 5, 0.239600660256161e2},
        {-1, 0, 0.122261479925384e-1},
        {-1, 2, 0.216356057692938e1},
        {0, 0, 0.398198903368642},
        {0, 1, -0.116892827834085},
        {1, 0, -0.102845919373532},
        {1, 2, -0.492676637589284},
        {2, 0, 0.655540456406790e-1},
        {3, 2, -0.240462535078530},
        {4, 0, -0.269798180310075e-1},
        {4, 1, 0.128369435967012},
        };

        static RegionResidualElement Region3Cdata[] = {
        {-12, 6, 3.11967788763030},
        {-12, 8, 2.76713458847564e+04},
        {-12, 10, 3.22583103403269e+07},
        {-10, 6, -3.42416065095363e+02},
        {-10, 8, -8.99732529907377e+05},
        {-10, 10, -7.93892049821251e+07},
        {-8, 5, 9.53193003217388e+01},
        {-8, 6, 2.29784742345072e+03},
        {-8, 7, 1.75336675322499e+05},
        {-6, 8, 7.91214365222792e+06},
        {-5, 1, 3.19933345844209e-05},
        {-5, 4, -6.59508863555767e+01},
        {-5, 7, -8.33426563212851e+05},
        {-4, 2, 6.45734680583292e-02},
        {-4, 8, -3.82031020570813e+06},
        {-3, 0, 4.06398848470079e-05},
        {-3, 3, 3.10327498492008e+01},
        {-2, 0, -8.92996718483724e-04},
        {-2, 4, 2.34604891591616e+02},
        {-2, 5, 3.77515668966951e+03},
        {-1, 0, 1.58646812591361e-02},
        {-1, 1, 7.07906336241843e-01},
        {-1, 2, 1.26016225146570e+01},
        {0, 0, 7.36143655772152e-01},
        {0, 1, 6.76544268999101e-01},
        {0, 2, -1.78100588189137e+01},
        {1, 0, -1.56531975531713e-01},
        {1, 2, 1.17707430048158e+01},
        {2, 0, 8.40143653860447e-02},
        {2, 1, -1.86442467471949e-01},
        {2, 3, -4.40170203949645e+01},
        {2, 7, 1.23290423502494e+06},
        {3, 0, -2.40650039730845e-02},
        {3, 7, -1.07077716660869e+06},
        {8, 1, 4.38319858566475e-02},
        };

        static RegionResidualElement Region3Ddata[] = {
        {-12, 4, -4.52484847171645e-10},
        {-12, 6, 3.15210389538801e-05},
        {-12, 7, -2.14991352047545e-03},
        {-12, 10, 5.08058874808345e+02},
        {-12, 12, -1.27123036845932e+07},
        {-12, 16, 1.15371133120497e+12},
        {-10, 0, -1.97805728776273e-16},
        {-10, 2, 2.41554806033972e-11},
        {-10, 4, -1.56481703640525e-06},
        {-10, 6, 2.77211346836625e-03},
        {-10, 8, -2.03578994462286e+01},
        {-10, 10, 1.44369489909053e+06},
        {-10, 14, -4.11254217946539e+10},
        {-8, 3, 6.23449786243773e-06},
        {-8, 7, -2.21774281146038e+01},
        {-8, 8, -6.89315087933158e+04},
        {-8, 10, -1.95419525060713e+07},
        {-6, 6, 3.16373510564015e+03},
        {-6, 8, 2.24040754426988e+06},
        {-5, 1, -4.36701347922356e-06},
        {-5, 2, -4.04213852833996e-04},
        {-5, 5, -3.48153203414663e+02},
        {-5, 7, -3.85294213555289e+05},
        {-4, 0, 1.35203700099403e-07},
        {-4, 1, 1.34648383271089e-04},
        {-4, 7, 1.25031835351736e+05},
        {-3, 2, 9.68123678455841e-02},
        {-3, 4, 2.25660517512438e+02},
        {-2, 0, -1.90102435341872e-04},
        {-2, 1, -2.99628410819229e-02},
        {-1, 0, 5.00833915372121e-03},
        {-1, 1, 3.87842482998411e-01},
        {-1, 5, -1.38535367777182e+03},
        {0, 0, 8.70745245971773e-01},
        {0, 2, 1.71946252068742},
        {1, 0, -3.26650121426383e-02},
        {1, 6, 4.98044171727877e+03},
        {3, 0, 5.51478022765087e-03},
        };

        static RegionResidualElement Region3Edata[] = {
        {-12, 14, 7.15815808404721e+08},
        {-12, 16, -1.14328360753449e+11},
        {-10, 3, 3.76531002015720e-12},
        {-10, 6, -9.03983668691157e-05},
        {-10, 10, 6.65695908836252e+05},
        {-10, 14, 5.35364174960127e+09},
        {-10, 16, 7.94977402335603e+10},
        {-8, 7, 9.22230563421437e+01},
        {-8, 8, -1.42586073991215e+05},
        {-8, 10, -1.11796381424162e+06},
        {-6, 6, 8.96121629640760e+03},
        {-5, 6, -6.69989239070491e+03},
        {-4, 2, 4.51242538486834e-03},
        {-4, 4, -3.39731325977713e+01},
        {-3, 2, -1.20523111552278},
        {-3, 6, 4.75992667717124e+04},
        {-3, 7, -2.66627750390341e+05},
        {-2, 0, -1.53314954386524e-04},
        {-2, 1, 3.05638404828265e-01},
        {-2, 3, 1.23654999499486e+02},
        {-2, 4, -1.04390794213011e+03},
        {-1, 0, -1.57496516174308e-02},
        {0, 0, 6.85331118940253e-01},
        {0, 1, 1.78373462873903},
        {1, 0, -5.44674124878910e-01},
        {1, 4, 2.04529931318843e+03},
        {1, 6, -2.28342359328752e+04},
        {2, 0, 4.13197481515899e-01},
        {2, 2, -3.41931835910405e+01},
        };

        static RegionResidualElement Region3Fdata[] = {
        {0, -3, -2.51756547792325e-08},
        {0, -2, 6.01307193668763e-06},
        {0, -1, -1.00615977450049e-03},
        {0, 0, 9.99969140252192e-01},
        {0, 1, 2.14107759236486},
        {0, 2, -1.65175571959086e+01},
        {1, -1, -1.41987303638727e-03},
        {1, 1, 2.69251915156554},
        {1, 2, 3.49741815858722e+01},
        {1, 3, -3.00208695771783e+01},
        {2, 0, -1.31546288252539},
        {2, 1, -8.39091277286169},
        {3, -5, 1.81545608337015e-10},
        {3, -2, -5.91099206478909e-04},
        {3, 0, 1.52115067087106},
        {4, -3, 2.52956470663225e-05},
        {5, -8, 1.00726265203786e-15},
        {5, 1, -1.49774533860650},
        {6, -6, -7.93940970562969e-10},
        {7, -4, -1.50290891264717e-04},
        {7, 1, 1.51205531275133},
        {10, -6, 4.70942606221652e-06},
        {12, -10, 1.95049710391712e-13},
        {12, -8, -9.11627886266077e-09},
        {12, -4, 6.04374640201265e-04},
        {14, -12, -2.25132933900136e-16},
        {14, -10, 6.10916973582981e-12},
        {14, -8, -3.03063908043404e-07},
        {14, -6, -1.37796070798409e-05},
        {14, -4, -9.19296736666106e-04},
        {16, -10, 6.39288223132545e-10},
        {16, -8, 7.53259479898699e-07},
        {18, -12, -4.00321478682929e-13},
        {18, -10, 7.56140294351614e-09},
        {20, -12, -9.12082054034891e-12},
        {20, -10, -2.37612381140539e-08},
        {20, -6, 2.69586010591874e-05},
        {22, -12, -7.32828135157839e-11},
        {24, -12, 2.41995578306660e-10},
        {24, -4, -4.05735532730322e-04},
        {28, -12, 1.89424143498011e-10},
        {32, -12, -4.86632965074563e-10},
        };

        static RegionResidualElement Region3Gdata[] = {
        {-12, 7, 4.12209020652996e-05},
        {-12, 12, -1.14987238280587e+06},
        {-12, 14, 9.48180885032080e+09},
        {-12, 18, -1.95788865718971e+17},
        {-12, 22, 4.96250704871300e+24},
        {-12, 24, -1.05549884548496e+28},
        {-10, 14, -7.58642165988278e+11},
        {-10, 20, -9.22172769596101e+22},
        {-10, 24, 7.25379072059348e+29},
        {-8, 7, -6.17718249205859e+01},
        {-8, 8, 1.07555033344858e+04},
        {-8, 10, -3.79545802336487e+07},
        {-8, 12, 2.28646846221831e+11},
        {-6, 8, -4.99741093010619e+06},
        {-6, 22, -2.80214310054101e+30},
        {-5, 7, 1.04915406769586e+06},
        {-5, 20, 6.13754229168619e+27},
        {-4, 22, 8.02056715528378e+31},
        {-3, 7, -2.98617819828065e+07},
        {-2, 3, -9.10782540134681e+01},
        {-2, 5, 1.35033227281565e+05},
        {-2, 14, -7.12949383408211e+18},
        {-2, 24, -1.04578785289542e+36},
        {-1, 2, 3.04331584444093e+01},
        {-1, 8, 5.93250797959445e+09},
        {-1, 18, -3.64174062110798e+27},
        {0, 0, 9.21791403532461e-01},
        {0, 1, -3.37693609657471e-01},
        {0, 2, -7.24644143758508e+01},
        {1, 0, -1.10480239272601e-01},
        {1, 1, 5.36516031875059},
        {1, 3, -2.91441872156205e+03},
        {3, 24, 6.16338176535305e+39},
        {5, 22, -1.20889175861180e+38},
        {6, 12, 8.18396024524612e+22},
        {8, 3, 9.40781944835829e+08},
        {10, 0, -3.67279669545448e+04},
        {10, 6, -8.37513931798655e+15},
        };

        static RegionResidualElement Region3Hdata[] = {
        {-12, 8, 5.61379678887577e-02},
        {-12, 12, 7.74135421587083e+09},
        {-10, 4, 1.11482975877938e-09},
        {-10, 6, -1.43987128208183e-03},
        {-10, 8, 1.93696558764920e+03},
        {-10, 10, -6.05971823585005e+08},
        {-10, 14, 1.71951568124337e+13},
        {-10, 16, -1.85461154985145e+16},
        {-8, 0, 3.87851168078010e-17},
        {-8, 1, -3.95464327846105e-14},
        {-8, 6, -1.70875935679023e+02},
        {-8, 7, -2.12010620701220e+03},
        {-8, 8, 1.77683337348191e+07},
        {-6, 4, 1.10177443629575e+01},
        {-6, 6, -2.34396091693313e+05},
        {-6, 8, -6.56174421999594e+06},
        {-5, 2, 1.56362212977396e-05},
        {-5, 3, -2.12946257021400},
        {-5, 4, 1.35249306374858e+01},
        {-4, 2, 1.77189164145813e-01},
        {-4, 4, 1.39499167345464e+03},
        {-3, 1, -7.03670932036388e-03},
        {-3, 2, -1.52011044389648e-01},
        {-2, 0, 9.81916922991113e-05},
        {-1, 0, 1.47199658618076e-03},
        {-1, 2, 2.02618487025578e+01},
        {0, 0, 8.99345518944240e-01},
        {1, 0, -2.11346402240858e-01},
        {1, 2, 2.49971752957491e+01},
        };

        static RegionResidualElement Region3Idata[] = {
        {0, 0, 1.06905684359136},
        {0, 1, -1.48620857922333},
        {0, 10, 2.59862256980408e+14},
        {1, -4, -4.46352055678749e-12},
        {1, -2, -5.66620757170032e-07},
        {1, -1, -2.35302885736849e-03},
        {1, 0, -2.69226321968839e-01},
        {2, 0, 9.22024992944392},
        {3, -5, 3.57633505503772e-12},
        {3, 0, -1.73942565562222e+01},
        {4, -3, 7.00681785556229e-06},
        {4, -2, -2.67050351075768e-04},
        {4, -1, -2.31779669675624},
        {5, -6, -7.53533046979752e-13},
        {5, -1, 4.81337131452891},
        {5, 12, -2.23286270422356e+21},
        {7, -4, -1.18746004987383e-05},
        {7, -3, 6.46412934136496e-03},
        {8, -6, -4.10588536330937e-10},
        {8, 10, 4.22739537057241e+19},
        {10, -8, 3.13698180473812e-13},
        {12, -12, 1.64395334345040e-24},
        {12, -6, -3.39823323754373e-06},
        {12, -4, -1.35268639905021e-02},
        {14, -10, -7.23252514211625e-15},
        {14, -8, 1.84386437538366e-09},
        {14, -4, -4.63959533752385e-02},
        {14, 5, -9.92263100376750e+13},
        {18, -12, 6.88169154439335e-17},
        {18, -10, -2.22620998452197e-11},
        {18, -8, -5.40843018624083e-08},
        {18, -6, 3.45570606200257e-03},
        {18, 2, 4.22275800304086e+10},
        {20, -12, -1.26974478770487e-15},
        {20, -10, 9.27237985153679e-10},
        {22, -12, 6.12670812016489e-14},
        {24, -12, -7.22693924063497e-12},
        {24, -8, -3.83669502636822e-04},
        {32, -10, 3.74684572410204e-04},
        {32, -5, -9.31976897511086e+04},
        {36, -10, -2.47690616026922e-02},
        {36, -8, 6.58110546759474e+01},
        };

        static RegionResidualElement Region3Jdata[] = {
        {0, -1, -1.11371317395540e-04},
        {0, 0, 1.00342892423685},
        {0, 1, 5.30615581928979},
        {1, -2, 1.79058760078792e-06},
        {1, -1, -7.28541958464774e-04},
        {1, 1, -1.87576133371704e+01},
        {2, -1, 1.99060874071849e-03},
        {2, 1, 2.43574755377290e+01},
        {3, -2, -1.77040785499444e-04},
        {4, -2, -2.59680385227130e-03},
        {4, 2, -1.98704578406823e+02},
        {5, -3, 7.38627790224287e-05},
        {5, -2, -2.36264692844138e-03},
        {5, 0, -1.61023121314333},
        {6, 3, 6.22322971786473e+03},
        {10, -6, -9.60754116701669e-09},
        {12, -8, -5.10572269720488e-11},
        {12, -3, 7.67373781404211e-03},
        {14, -10, 6.63855469485254e-15},
        {14, -8, -7.17590735526745e-10},
        {14, -5, 1.46564542926508e-05},
        {16, -10, 3.09029474277013e-12},
        {18, -12, -4.64216300971708e-16},
        {20, -12, -3.90499637961161e-14},
        {20, -10, -2.36716126781431e-10},
        {24, -12, 4.54652854268717e-12},
        {24, -6, -4.22271787482497e-03},
        {28, -12, 2.83911742354706e-11},
        {28, -5, 2.70929002720228},
        };

        static RegionResidualElement Region3Kdata[] = {
        {-2, 10, -4.01215699576099e+08},
        {-2, 12, 4.84501478318406e+10},
        {-1, -5, 3.94721471363678e-15},
        {-1, 6, 3.72629967374147e+04},
        {0, -12, -3.69794374168666e-30},
        {0, -6, -3.80436407012452e-15},
        {0, -2, 4.75361629970233e-07},
        {0, -1, -8.79148916140706e-04},
        {0, 0, 8.44317863844331e-01},
        {0, 1, 1.22433162656600e+01},
        {0, 2, -1.04529634830279e+02},
        {0, 3, 5.89702771277429e+02},
        {0, 14, -2.91026851164444e+13},
        {1, -3, 1.70343072841850e-06},
        {1, -2, -2.77617606975748e-04},
        {1, 0, -3.44709605486686},
        {1, 1, 2.21333862447095e+01},
        {1, 2, -1.94646110037079e+02},
        {2, -8, 8.08354639772825e-16},
        {2, -6, -1.80845209145470e-11},
        {2, -3, -6.96664158132412e-06},
        {2, -2, -1.81057560300994e-03},
        {2, 0, 2.55830298579027},
        {2, 4, 3.28913873658481e+03},
        {5, -12, -1.73270241249904e-19},
        {5, -6, -6.61876792558034e-07},
        {5, -3, -3.95688923421250e-03},
        {6, -12, 6.04203299819132e-18},
        {6, -10, -4.00879935920517e-14},
        {6, -8, 1.60751107464958e-09},
        {6, -5, 3.83719409025556e-05},
        {8, -12, -6.49565446702457e-15},
        {10, -12, -1.49095328506000e-12},
        {12, -10, 5.41449377329581e-09},
        };

        static RegionResidualElement Region3Ldata[] = {
        {-12, 14, 2.60702058647537e+09},
        {-12, 16, -1.88277213604704e+14},
        {-12, 18, 5.54923870289667e+18},
        {-12, 20, -7.58966946387758e+22},
        {-12, 22, 4.13865186848908e+26},
        {-10, 14, -8.15038000738060e+11},
        {-10, 24, -3.81458260489955e+32},
        {-8, 6, -1.23239564600519e-02},
        {-8, 10, 2.26095631437174e+07},
        {-8, 12, -4.95017809506720e+11},
        {-8, 14, 5.29482996422863e+15},
        {-8, 18, -4.44359478746295e+22},
        {-8, 24, 5.21635864527315e+34},
        {-8, 36, -4.87095672740742e+54},
        {-6, 8, -7.14430209937547e+05},
        {-5, 4, 1.27868634615495e-01},
        {-5, 5, -1.00752127917598e+01},
        {-4, 7, 7.77451437960990e+06},
        {-4, 16, -1.08105480796471e+24},
        {-3, 1, -3.57578581169659e-06},
        {-3, 3, -2.12857169423484},
        {-3, 18, 2.70706111085238e+29},
        {-3, 20, -6.95953622348829e+32},
        {-2, 2, 1.10609027472280e-01},
        {-2, 3, 7.21559163361354e+01},
        {-2, 10, -3.06367307532219e+14},
        {-1, 0, 2.65839618885530e-05},
        {-1, 1, 2.53392392889754e-02},
        {-1, 3, -2.14443041836579e+02},
        {0, 0, 9.37846601489667e-01},
        {0, 1, 2.23184043101700},
        {0, 2, 3.38401222509191e+01},
        {0, 12, 4.94237237179718e+20},
        {1, 0, -1.98068404154428e-01},
        {1, 16, -1.41415349881140e+30},
        {2, 1, -9.93862421613651e+01},
        {4, 0, 1.25070534142731e+02},
        {5, 0, -9.96473529004439e+02},
        {5, 1, 4.73137909872765e+04},
        {6, 14, 1.16662121219322e+32},
        {10, 4, -3.15874976271533e+15},
        {10, 12, -4.45703369196945e+32},
        {14, 10, 6.42794932373694e+32},
        };

        static RegionResidualElement Region3Mdata[] = {
        {0, 0, 8.11384363481847e-01},
        {3, 0, -5.68199310990094e+03},
        {8, 0, -1.78657198172556e+10},
        {20, 2, 7.95537657613427e+31},
        {1, 5, -8.14568209346872e+04},
        {3, 5, -6.59774567602874e+07},
        {4, 5, -1.52861148659302e+10},
        {5, 5, -5.60165667510446e+11},
        {1, 6, 4.58384828593949e+05},
        {6, 6, -3.85754000383848e+13},
        {2, 7, 4.53735800004273e+07},
        {4, 8, 9.39454935735563e+11},
        {14, 8, 2.66572856432938e+27},
        {2, 10, -5.47578313899097e+09},
        {5, 10, 2.00725701112386e+14},
        {3, 12, 1.85007245563239e+12},
        {0, 14, 1.85135446828337e+08},
        {1, 14, -1.70451090076385e+11},
        {1, 18, 1.57890366037614e+14},
        {1, 20, -2.02530509748774e+15},
        {28, 20, 3.68193926183570e+59},
        {2, 22, 1.70215539458936e+17},
        {16, 22, 6.39234909918741e+41},
        {0, 24, -8.21698160721956e+14},
        {5, 24, -7.95260241872306e+23},
        {0, 28, 2.33415869478510e+17},
        {3, 28, -6.00079934586803e+22},
        {4, 28, 5.94584382273384e+24},
        {12, 28, 1.89461279349492e+39},
        {16, 28, -8.10093428842645e+45},
        {1, 32, 1.88813911076809e+21},
        {8, 32, 1.11052244098768e+35},
        {14, 32, 2.91133958602503e+45},
        {0, 36, -3.29421923951460e+21},
        {2, 36, -1.37570282536696e+25},
        {3, 36, 1.81508996303902e+27},
        {4, 36, -3.46865122768353e+29},
        {8, 36, -2.11961148774260e+37},
        {14, 36, -1.28617899887675e+48},
        {24, 36, 4.79817895699239e+64},
        };

        static RegionResidualElement Region3Ndata[] = {
        {0, -12, 2.80967799943151e-39},
        {3, -12, 6.14869006573609e-31},
        {4, -12, 5.82238667048942e-28},
        {6, -12, 3.90628369238462e-23},
        {7, -12, 8.21445758255119e-21},
        {10, -12, 4.02137961842776e-15},
        {12, -12, 6.51718171878301e-13},
        {14, -12, -2.11773355803058e-08},
        {18, -12, 2.64953354380072e-03},
        {0, -10, -1.35031446451331e-32},
        {3, -10, -6.07246643970893e-24},
        {5, -10, -4.02352115234494e-19},
        {6, -10, -7.44938506925544e-17},
        {8, -10, 1.89917206526237e-13},
        {12, -10, 3.64975183508473e-06},
        {0, -8, 1.77274872361946e-26},
        {3, -8, -3.34952758812999e-19},
        {7, -8, -4.21537726098389e-09},
        {12, -8, -3.91048167929649e-02},
        {2, -6, 5.41276911564176e-14},
        {3, -6, 7.05412100773699e-12},
        {4, -6, 2.58585887897486e-09},
        {2, -5, -4.93111362030162e-11},
        {4, -5, -1.58649699894543e-06},
        {7, -5, -5.25037427886100e-01},
        {4, -4, 2.20019901729615e-03},
        {3, -3, -6.43064132636925e-03},
        {5, -3, 6.29154149015048e+01},
        {6, -3, 1.35147318617061e+02},
        {0, -2, 2.40560808321713e-07},
        {0, -1, -8.90763306701305e-04},
        {3, -1, -4.40209599407714e+03},
        {1, 0, -3.02807107747776e+02},
        {0, 1, 1.59158748314599e+03},
        {1, 1, 2.32534272709876e+05},
        {0, 2, -7.92681207132600e+05},
        {1, 4, -8.69871364662769e+10},
        {0, 5, 3.54542769185671e+11},
        {1, 6, 4.00849240129329e+14},
        };

        static RegionResidualElement Region3Odata[] = {
        {0, -12, 1.28746023979718e-35},
        {0, -4, -7.35234770382342e-12},
        {0, -1, 2.89078692149150e-03},
        {2, -1, 2.44482731907223e-01},
        {3, -10, 1.41733492030985e-24},
        {4, -12, -3.54533853059476e-29},
        {4, -8, -5.94539202901431e-18},
        {4, -5, -5.85188401782779e-09},
        {4, -4, 2.01377325411803e-06},
        {4, -1, 1.38647388209306},
        {5, -4, -1.73959365084772e-05},
        {5, -3, 1.37680878349369e-03},
        {6, -8, 8.14897605805513e-15},
        {7, -12, 4.25596631351839e-26},
        {8, -10, -3.87449113787755e-18},
        {8, -8, 1.39814747930240e-13},
        {8, -4, -1.71849638951521e-03},
        {10, -12, 6.41890529513296e-22},
        {10, -8, 1.18960578072018e-11},
        {14, -12, -1.55282762571611e-18},
        {14, -8, 2.33907907347507e-08},
        {20, -12, -1.74093247766213e-13},
        {20, -10, 3.77682649089149e-09},
        {24, -12, -5.16720236575302e-11},
        };

        static RegionResidualElement Region3Pdata[] = {
        {0, -1, -9.82825342010366e-05},
        {0, 0, 1.05145700850612},
        {0, 1, 1.16033094095084e+02},
        {0, 2, 3.24664750281543e+03},
        {1, 1, -1.23592348610137e+03},
        {2, -1, -5.61403450013495e-02},
        {3, -3, 8.56677401640869e-08},
        {3, 0, 2.36313425393924e+02},
        {4, -2, 9.72503292350109e-03},
        {6, -2, -1.03001994531927},
        {7, -5, -1.49653706199162e-09},
        {7, -4, -2.15743778861592e-05},
        {8, -2, -8.34452198291445},
        {10, -3, 5.86602660564988e-01},
        {12, -12, 3.43480022104968e-26},
        {12, -6, 8.16256095947021e-06},
        {12, -5, 2.94985697916798e-03},
        {14, -10, 7.11730466276584e-17},
        {14, -8, 4.00954763806941e-10},
        {14, -3, 1.07766027032853e+01},
        {16, -8, -4.09449599138182e-07},
        {18, -8, -7.29121307758902e-06},
        {20, -10, 6.77107970938909e-09},
        {22, -10, 6.02745973022975e-08},
        {24, -12, -3.82323011855257e-11},
        {24, -8, 1.79946628317437e-03},
        {36, -12, -3.45042834640005e-04},
        };

        static RegionResidualElement Region3Qdata[] = {
        {-12, 10, -8.20433843259950e+04},
        {-12, 12, 4.73271518461586e+10},
        {-10, 6, -8.05950021005413e-02},
        {-10, 7, 3.28600025435980e+01},
        {-10, 8, -3.56617029982490e+03},
        {-10, 10, -1.72985781433335e+09},
        {-8, 8, 3.51769232729192e+07},
        {-6, 6, -7.75489259985144e+05},
        {-5, 2, 7.10346691966018e-05},
        {-5, 5, 9.93499883820274e+04},
        {-4, 3, -6.42094171904570e-01},
        {-4, 4, -6.12842816820083e+03},
        {-3, 3, 2.32808472983776e+02},
        {-2, 0, -1.42808220416837e-05},
        {-2, 1, -6.43596060678456e-03},
        {-2, 2, -4.28577227475614},
        {-2, 4, 2.25689939161918e+03},
        {-1, 0, 1.00355651721510e-03},
        {-1, 1, 3.33491455143516e-01},
        {-1, 2, 1.09697576888873},
        {0, 0, 9.61917379376452e-01},
        {1, 0, -8.38165632204598e-02},
        {1, 1, 2.47795908411492},
        {1, 3, -3.19114969006533e+03},
        };

        static RegionResidualElement Region3Rdata[] = {
        {-8, 6, 1.44165955660863e-03},
        {-8, 14, -7.01438599628258e+12},
        {-3, -3, -8.30946716459219e-17},
        {-3, 3, 2.61975135368109e-01},
        {-3, 4, 3.93097214706245e+02},
        {-3, 5, -1.04334030654021e+04},
        {-3, 8, 4.90112654154211e+08},
        {0, -1, -1.47104222772069e-04},
        {0, 0, 1.03602748043408},
        {0, 1, 3.05308890065089},
        {0, 5, -3.99745276971264e+06},
        {3, -6, 5.69233719593750e-12},
        {3, -2, -4.64923504407778e-02},
        {8, -12, -5.35400396512906e-18},
        {8, -10, 3.99988795693162e-13},
        {8, -8, -5.36479560201811e-07},
        {8, -5, 1.59536722411202e-02},
        {10, -12, 2.70303248860217e-15},
        {10, -10, 2.44247453858506e-08},
        {10, -8, -9.83430636716454e-06},
        {10, -6, 6.63513144224454e-02},
        {10, -5, -9.93456957845006},
        {10, -4, 5.46491323528491e+02},
        {10, -3, -1.43365406393758e+04},
        {10, -2, 1.50764974125511e+05},
        {12, -12, -3.37209709340105e-10},
        {14, -12, 3.77501980025469e-09},
        };

        static RegionResidualElement Region3Sdata[] = {
        {-12, 20, -5.32466612140254e+22},
        {-12, 24, 1.00415480000824e+31},
        {-10, 22, -1.91540001821367e+29},
        {-8, 14, 1.05618377808847e+16},
        {-6, 36, 2.02281884477061e+58},
        {-5, 8, 8.84585472596134e+07},
        {-5, 16, 1.66540181638363e+22},
        {-4, 6, -3.13563197669111e+05},
        {-4, 32, -1.85662327545324e+53},
        {-3, 3, -6.24942093918942e-02},
        {-3, 8, -5.04160724132590e+09},
        {-2, 4, 1.87514491833092e+04},
        {-1, 1, 1.21399979993217e-03},
        {-1, 2, 1.88317043049455},
        {-1, 3, -1.67073503962060e+03},
        {0, 0, 9.65961650599775e-01},
        {0, 1, 2.94885696802488},
        {0, 4, -6.53915627346115e+04},
        {0, 28, 6.04012200163444e+49},
        {1, 0, -1.98339358557937e-01},
        {1, 32, -1.75984090163501e+57},
        {3, 0, 3.56314881403987},
        {3, 1, -5.75991255144384e+02},
        {3, 2, 4.56213415338071e+04},
        {4, 3, -1.09174044987829e+07},
        {4, 18, 4.37796099975134e+33},
        {4, 24, -6.16552611135792e+45},
        {5, 4, 1.93568768917797e+09},
        {14, 24, 9.50898170425042e+53},
        };

        static RegionResidualElement Region3Tdata[] = {
        {0, 0, 1.55287249586268},
        {0, 1, 6.64235115009031},
        {0, 4, -2.89366236727210e+03},
        {0, 12, -3.85923202309848e+12},
        {1, 0, -2.91002915783761},
        {1, 10, -8.29088246858083e+11},
        {2, 0, 1.76814899675218},
        {2, 6, -5.34686695713469e+08},
        {2, 14, 1.60464608687834e+17},
        {3, 3, 1.96435366560186e+05},
        {3, 8, 1.56637427541729e+12},
        {4, 0, -1.78154560260006},
        {4, 10, -2.29746237623692e+15},
        {7, 3, 3.85659001648006e+07},
        {7, 4, 1.10554446790543e+09},
        {7, 7, -6.77073830687349e+13},
        {7, 20, -3.27910592086523e+30},
        {7, 36, -3.41552040860644e+50},
        {10, 10, -5.27251339709047e+20},
        {10, 12, 2.45375640937055e+23},
        {10, 14, -1.68776617209269e+26},
        {10, 16, 3.58958955867578e+28},
        {10, 22, -6.56475280339411e+35},
        {18, 18, 3.55286045512301e+38},
        {20, 32, 5.69021454413270e+57},
        {22, 22, -7.00584546433113e+47},
        {22, 36, -7.05772623326374e+64},
        {24, 24, 1.66861176200148e+52},
        {28, 28, -3.00475129680486e+60},
        {32, 22, -6.68481295196808e+50},
        {32, 32, 4.28432338620678e+68},
        {32, 36, -4.44227367758304e+71},
        {36, 36, -2.81396013562745e+76},
        };

        static RegionResidualElement Region3Udata[] = {
        {-12, 14, 1.22088349258355e+17},
        {-10, 10, 1.04216468608488e+09},
        {-10, 12, -8.82666931564652e+15},
        {-10, 14, 2.59929510849499e+19},
        {-8, 10, 2.22612779142211e+14},
        {-8, 12, -8.78473585050085e+17},
        {-8, 14, -3.14432577551552e+21},
        {-6, 8, -2.16934916996285e+12},
        {-6, 12, 1.59079648196849e+20},
        {-5, 4, -3.39567617303423e+02},
        {-5, 8, 8.84387651337836e+12},
        {-5, 12, -8.43405926846418e+20},
        {-3, 2, 1.14178193518022e+01},
        {-1, -1, -1.22708229235641e-04},
        {-1, 1, -1.06201671767107e+02},
        {-1, 12, 9.03443213959313e+24},
        {-1, 14, -6.93996270370852e+27},
        {0, -3, 6.48916718965575e-09},
        {0, 1, 7.18957567127851e+03},
        {1, -2, 1.05581745346187e-03},
        {2, 5, -6.51903203602581e+14},
        {2, 10, -1.60116813274676e+24},
        {3, -5, -5.10254294237837e-09},
        {5, -4, -1.52355388953402e-01},
        {5, 2, 6.77143292290144e+11},
        {5, 3, 2.76378438378930e+14},
        {6, -5, 1.16862983141686e-02},
        {6, 2, -3.01426947980171e+13},
        {8, -8, 1.69719813884840e-08},
        {8, 8, 1.04674840020929e+26},
        {10, -4, -1.08016904560140e+04},
        {12, -12, -9.90623601934295e-13},
        {12, -4, 5.36116483602738e+06},
        {12, 4, 2.26145963747881e+21},
        {14, -12, -4.88731565776210e-10},
        {14, -10, 1.51001548880670e-05},
        {14, -6, -2.27700464643920e+04},
        {14, 6, -7.81754507698846e+27},
        };

        static RegionResidualElement Region3Vdata[] = {
        {-10, -8, -4.15652812061591e-55},
        {-8, -12, 1.77441742924043e-61},
        {-6, -12, -3.57078668203377e-55},
        {-6, -3, 3.59252213604114e-26},
        {-6, 5, -2.59123736380269e+01},
        {-6, 6, 5.94619766193460e+04},
        {-6, 8, -6.24184007103158e+10},
        {-6, 10, 3.13080299915944e+16},
        {-5, 1, 1.05006446192036e-09},
        {-5, 2, -1.92824336984852e-06},
        {-5, 6, 6.54144373749937e+05},
        {-5, 8, 5.13117462865044e+12},
        {-5, 10, -6.97595750347391e+18},
        {-5, 14, -1.03977184454767e+28},
        {-4, -12, 1.19563135540666e-48},
        {-4, -10, -4.36677034051655e-42},
        {-4, -6, 9.26990036530639e-30},
        {-4, 10, 5.87793105620748e+20},
        {-3, -3, 2.80375725094731e-18},
        {-3, 10, -1.92359972440634e+22},
        {-3, 12, 7.42705723302738e+26},
        {-2, 2, -5.17429682450605e+01},
        {-2, 4, 8.20612048645469e+06},
        {-1, -2, -1.88214882341448e-09},
        {-1, 0, 1.84587261114837e-02},
        {0, -2, -1.35830407782663e-06},
        {0, 6, -7.23681885626348e+16},
        {0, 10, -2.23449194054124e+26},
        {1, -12, -1.11526741826431e-35},
        {1, -10, 2.76032601145151e-29},
        {3, 3, 1.34856491567853e+14},
        {4, -6, 6.52440293345860e-10},
        {4, 3, 5.10655119774360e+16},
        {4, 10, -4.68138358908732e+31},
        {5, 2, -7.60667491183279e+15},
        {8, -12, -4.17247986986821e-19},
        {10, -2, 3.12545677756104e+13},
        {12, -3, -1.00375333864186e+14},
        {14, 1, 2.47761392329058e+26},
        };

        static RegionResidualElement Region3Wdata[] = {
        {-12, 8, -5.86219133817016e-08},
        {-12, 14, -8.94460355005526e+10},
        {-10, -1, 5.31168037519774e-31},
        {-10, 8, 1.09892402329239e-01},
        {-8, 6, -5.75368389425212e-02},
        {-8, 8, 2.28276853990249e+04},
        {-8, 14, -1.58548609655002e+18},
        {-6, -4, 3.29865748576503e-28},
        {-6, -3, -6.34987981190669e-25},
        {-6, 2, 6.15762068640611e-09},
        {-6, 8, -9.61109240985747e+07},
        {-5, -10, -4.06274286652625e-45},
        {-4, -1, -4.71103725498077e-13},
        {-4, 3, 7.25937724828145e-01},
        {-3, -10, 1.87768525763682e-39},
        {-3, 3, -1.03308436323771e+03},
        {-2, 1, -6.62552816342168e-02},
        {-2, 2, 5.79514041765710e+02},
        {-1, -8, 2.37416732616644e-27},
        {-1, -4, 2.71700235739893e-15},
        {-1, 1, -9.07886213483600e+01},
        {0, -12, -1.71242509570207e-37},
        {0, 1, 1.56792067854621e+02},
        {1, -1, 9.23261357901470e-01},
        {2, -1, -5.97865988422577},
        {2, 2, 3.21988767636389e+06},
        {3, -12, -3.99441390042203e-30},
        {3, -5, 4.93429086046981e-08},
        {5, -10, 8.12036983370565e-20},
        {5, -8, -2.07610284654137e-12},
        {5, -6, -3.40821291419719e-07},
        {8, -12, 5.42000573372233e-18},
        {8, -10, -8.56711586510214e-13},
        {10, -12, 2.66170454405981e-14},
        {10, -8, 8.58133791857099e-06},
        };

        static RegionResidualElement Region3Xdata[] = {
        {-8, 14, 3.77373741298151e+18},
        {-6, 10, -5.07100883722913e+12},
        {-5, 10, -1.03363225598860e+15},
        {-4, 1, 1.84790814320773e-06},
        {-4, 2, -9.24729378390945e-04},
        {-4, 14, -4.25999562292738e+23},
        {-3, -2, -4.62307771873973e-13},
        {-3, 12, 1.07319065855767e+21},
        {-1, 5, 6.48662492280682e+10},
        {0, 0, 2.44200600688281},
        {0, 4, -8.51535733484258e+09},
        {0, 10, 1.69894481433592e+21},
        {1, -10, 2.15780222509020e-27},
        {1, -1, -3.20850551367334e-01},
        {2, 6, -3.82642448458610e+16},
        {3, -12, -2.75386077674421e-29},
        {3, 0, -5.63199253391666e+05},
        {3, 8, -3.26068646279314e+20},
        {4, 3, 3.97949001553184e+13},
        {5, -6, 1.00824008584757e-07},
        {5, -2, 1.62234569738433e+04},
        {5, 1, -4.32355225319745e+10},
        {6, 1, -5.92874245598610e+11},
        {8, -6, 1.33061647281106},
        {8, -3, 1.57338197797544e+06},
        {8, 1, 2.58189614270853e+13},
        {8, 8, 2.62413209706358e+24},
        {10, -8, -9.20011937431142e-02},
        {12, -10, 2.20213765905426e-03},
        {12, -8, -1.10433759109547e+01},
        {12, -5, 8.47004870612087e+06},
        {12, -4, -5.92910695762536e+08},
        {14, -12, -1.83027173269660e-05},
        {14, -10, 1.81339603516302e-01},
        {14, -8, -1.19228759669889e+03},
        {14, -6, 4.30867658061468e+06},
        };

        static RegionResidualElement Region3Ydata[] = {
        {0, -3, -5.25597995024633e-10},
        {0, 1, 5.83441305228407e+03},
        {0, 5, -1.34778968457925e+16},
        {0, 8, 1.18973500934212e+25},
        {1, 8, -1.59096490904708e+26},
        {2, -4, -3.15839902302021e-07},
        {2, -1, 4.96212197158239e+02},
        {2, 4, 3.27777227273171e+18},
        {2, 5, -5.27114657850696e+21},
        {3, -8, 2.10017506281863e-17},
        {3, 4, 7.05106224399834e+20},
        {3, 8, -2.66713136106469e+30},
        {4, -6, -1.45370512554562e-08},
        {4, 6, 1.49333917053130e+27},
        {5, -2, -1.49795620287641e+07},
        {5, 1, -3.81881906271100e+15},
        {8, -8, 7.24660165585797e-05},
        {8, -2, -9.37808169550193e+13},
        {10, -5, 5.14411468376383e+09},
        {12, -8, -8.28198594040141e+04},
        };

        static RegionResidualElement Region3Zdata[] = {
        {-8, 3, 2.4400789229065e-11},
        {-6, 6, -4.6305743033124e+06},
        {-5, 6, 7.2880327477771e+09},
        {-5, 8, 3.2777630285886e+15},
        {-4, 5, -1.1059817011841e+09},
        {-4, 6, -3.2389991572996e+12},
        {-4, 8, 9.2381400702325e+15},
        {-3, -2, 8.4225008041371e-13},
        {-3, 5, 6.6322143624551e+11},
        {-3, 6, -1.6717018667214e+14},
        {-2, 2, 2.5374935870139e+03},
        {-1, -6, -8.1973155961052e-21},
        {0, 3, 3.2838058789066e+11},
        {1, 1, -6.2500479117154e+07},
        {2, 6, 8.0319795746202e+20},
        {3, -6, -2.0439701133835e-11},
        {3, -2, -3.7839104705594e+03},
        {6, -6, 9.7287654593862e-03},
        {6, -5, 1.5435572168146e+01},
        {6, -4, -3.7396286292864e+03},
        {6, -1, -6.8285901137457e+10},
        {8, -8, -2.4848801561454e-04},
        {8, -4, 3.9453604949707e+06},
        };

        class Region3BackwardsRegion{
        protected:
            double v_star, p_star, T_star;
            std::size_t N;
            double a, b, c, d, e;
            std::vector<int> I, J;
            std::vector<double> n;
        public:
    
            Region3BackwardsRegion(RegionResidualElement data[], std::size_t N){
                this->N = N;
                for (std::size_t i = 0; i < N; ++i){
                    n.push_back(data[i].n);
                    I.push_back(data[i].I);
                    J.push_back(data[i].J);
                }
            }
            virtual double v(double T, double p){
                const double pi = p/p_star, theta = T/T_star;
                double summer = 0;
                for (std::size_t i = 0; i < N; ++i){
                    summer += n[i]*pow(pow(pi-a, c), I[i])*pow(pow(theta-b, d), J[i]);
                }
                return pow(summer, e)*v_star;
            };
        };

        class Region3a : public Region3BackwardsRegion{
        public:
            Region3a() : Region3BackwardsRegion(Region3Adata, 30){ 
                v_star = 0.0024; p_star = 100e6; T_star = 760; a = 0.085; b = 0.817; c = 1; d = 1; e = 1;
            };
        };
        class Region3b : public Region3BackwardsRegion{
        public:
            Region3b() : Region3BackwardsRegion(Region3Bdata, 32){
                v_star = 0.0041; p_star = 100e6; T_star = 860; a = 0.280; b = 0.779; c = 1; d = 1; e = 1;
            };
        };
        class Region3c : public Region3BackwardsRegion{
        public:
            Region3c() : Region3BackwardsRegion(Region3Cdata, 35){
                v_star = 0.0022; p_star = 40e6; T_star = 690; a = 0.259; b = 0.903; c = 1; d = 1; e = 1;
            };
        };
        class Region3d : public Region3BackwardsRegion{
        public:
            Region3d() : Region3BackwardsRegion(Region3Ddata, 38){
                v_star = 0.0029; p_star = 40e6; T_star = 690; a = 0.559; b = 0.939; c = 1; d = 1; e = 4;
            };
        };
        class Region3e : public Region3BackwardsRegion{
        public:
            Region3e() : Region3BackwardsRegion(Region3Edata, 29){
                v_star = 0.0032; p_star = 40e6; T_star = 710; a = 0.587; b = 0.918; c = 1; d = 1; e = 1;
            };
        };
        class Region3f : public Region3BackwardsRegion{
        public:
            Region3f() : Region3BackwardsRegion(Region3Fdata, 42){
                v_star = 0.0064; p_star = 40e6; T_star = 730; a = 0.587; b = 0.891; c = 0.5; d = 1; e = 4;
            };
        };
        class Region3g : public Region3BackwardsRegion{
        public:
            Region3g() : Region3BackwardsRegion(Region3Gdata, 38){
                v_star = 0.0027; p_star = 25e6; T_star = 660; a = 0.872; b = 0.971; c = 1; d = 1; e = 4;
            };
        };
        class Region3h : public Region3BackwardsRegion{
        public:
            Region3h() : Region3BackwardsRegion(Region3Hdata, 29){
                v_star = 0.0032; p_star = 25e6; T_star = 660; a = 0.898; b = 0.983; c = 1; d = 1; e = 4;
            };
        };
        class Region3i : public Region3BackwardsRegion{
        public:
            Region3i() : Region3BackwardsRegion(Region3Idata, 42){
                v_star = 0.0041; p_star = 25e6; T_star = 660; a = 0.910; b = 0.984; c = 0.5; d = 1; e = 4;
            };
        };
        class Region3j : public Region3BackwardsRegion{
        public:
            Region3j() : Region3BackwardsRegion(Region3Jdata, 29){
                v_star = 0.0054; p_star = 25e6; T_star = 670; a = 0.875; b = 0.964; c = 0.5; d = 1; e = 4;
            };
        };
        class Region3k : public Region3BackwardsRegion{
        public:
            Region3k() : Region3BackwardsRegion(Region3Kdata, 34){
                v_star = 0.0077; p_star = 25e6; T_star = 680; a = 0.802; b = 0.935; c = 1; d = 1; e = 1;
            };
        };
        class Region3l : public Region3BackwardsRegion{
        public:
            Region3l() : Region3BackwardsRegion(Region3Ldata, 43){
                v_star = 0.0026; p_star = 24e6; T_star = 650; a = 0.908; b = 0.989; c = 1; d = 1; e = 4;
            };
        };
        class Region3m : public Region3BackwardsRegion{
        public:
            Region3m() : Region3BackwardsRegion(Region3Mdata, 40){
                v_star = 0.0028; p_star = 23e6; T_star = 650; a = 1.0; b = 0.997; c = 1; d = 0.25; e = 1;
            };
        };
        class Region3n : public Region3BackwardsRegion{
        public:
            Region3n() : Region3BackwardsRegion(Region3Ndata, 39){
                v_star = 0.0031; p_star = 23e6; T_star = 650; a = 0.976; b = 0.997;
            };
            double v(double T, double p){
                const double pi = p/p_star, theta = T/T_star;
                double summer = 0;
                for (std::size_t i = 0; i < N; ++i){
                    summer += n[i]*pow(pi-a, I[i])*pow(theta-b, J[i]);
                }
                return exp(summer)*v_star;
            };
        };
        class Region3o : public Region3BackwardsRegion{
        public:
            Region3o() : Region3BackwardsRegion(Region3Odata, 24){
                v_star = 0.0034; p_star = 23e6; T_star = 650; a = 0.974; b = 0.996; c = 0.5; d = 1; e = 1;
            };
        };
        class Region3p : public Region3BackwardsRegion{
        public:
            Region3p() : Region3BackwardsRegion(Region3Pdata, 27){
                v_star = 0.0041; p_star = 23e6; T_star = 650; a = 0.972; b = 0.997; c = 0.5; d = 1; e = 1;
            };
        };
        class Region3q : public Region3BackwardsRegion{
        public:
            Region3q() : Region3BackwardsRegion(Region3Qdata, 24){
                v_star = 0.0022; p_star = 23e6; T_star = 650; a = 0.848; b = 0.983; c = 1; d = 1; e = 4;
            };
        };
        class Region3r : public Region3BackwardsRegion{
        public:
            Region3r() : Region3BackwardsRegion(Region3Rdata, 27){
                v_star = 0.0054; p_star = 23e6; T_star = 650; a = 0.874; b = 0.982; c = 1; d = 1; e = 1;
            };
        };
        class Region3s : public Region3BackwardsRegion{
        public:
            Region3s() : Region3BackwardsRegion(Region3Sdata, 29){
                v_star = 0.0022; p_star = 21e6; T_star = 640; a = 0.886; b = 0.990; c = 1; d = 1; e = 4;
            };
        };
        class Region3t : public Region3BackwardsRegion{
        public:
            Region3t() : Region3BackwardsRegion(Region3Tdata, 33){
                v_star = 0.0088; p_star = 20e6; T_star = 650; a = 0.803; b = 1.02; c = 1; d = 1; e = 1;
            };
        };
        class Region3u : public Region3BackwardsRegion{
        public:
            Region3u() : Region3BackwardsRegion(Region3Udata, 38){
                v_star = 0.0026; p_star = 23e6; T_star = 650; a = 0.902; b = 0.988; c = 1; d = 1; e = 1;
            };
        };
        class Region3v : public Region3BackwardsRegion{
        public:
            Region3v() : Region3BackwardsRegion(Region3Vdata, 39){
                v_star = 0.0031; p_star = 23e6; T_star = 650; a = 0.960; b = 0.995; c = 1; d = 1; e = 1;
            };
        };
        class Region3w : public Region3BackwardsRegion{
        public:
            Region3w() : Region3BackwardsRegion(Region3Wdata, 35){
                v_star = 0.0039; p_star = 23e6; T_star = 650; a = 0.959; b = 0.995; c = 1; d = 1; e = 4;
            };
        };
        class Region3x : public Region3BackwardsRegion{
        public:
            Region3x() : Region3BackwardsRegion(Region3Xdata, 36){
                v_star = 0.0049; p_star = 23e6; T_star = 650; a = 0.910; b = 0.988; c = 1; d = 1; e = 1;
            };
        };
        class Region3y : public Region3BackwardsRegion{
        public:
            Region3y() : Region3BackwardsRegion(Region3Ydata, 20){
                v_star = 0.0031; p_star = 22e6; T_star = 650; a = 0.996; b = 0.994; c = 1; d = 1; e = 4;
            };
        };
        class Region3z : public Region3BackwardsRegion{
        public:
            Region3z() : Region3BackwardsRegion(Region3Zdata, 23){
                v_star = 0.0038; p_star = 22e6; T_star = 650; a = 0.993; b = 0.994; c = 1; d = 1; e = 4;
            };
        };

        inline double Region3_v_TP(char region, double T, double p){
            static Region3a R3a;
            static Region3b R3b;
            static Region3c R3c;
            static Region3d R3d;
            static Region3e R3e;
            static Region3f R3f;
            static Region3g R3g;
            static Region3h R3h;
            static Region3i R3i;
            static Region3j R3j;
            static Region3k R3k;
            static Region3l R3l;
            static Region3m R3m;
            static Region3n R3n;
            static Region3o R3o;
            static Region3p R3p;
            static Region3q R3q;
            static Region3r R3r;
            static Region3s R3s;
            static Region3t R3t;
            static Region3u R3u;
            static Region3v R3v;
            static Region3w R3w;
            static Region3x R3x;
            static Region3y R3y;
            static Region3z R3z;

            switch(region){
                case 'A': return R3a.v(T, p);
                case 'B': return R3b.v(T, p);
                case 'C': return R3c.v(T, p);
                case 'D': return R3d.v(T, p);
                case 'E': return R3e.v(T, p);
                case 'F': return R3f.v(T, p);
                case 'G': return R3g.v(T, p);
                case 'H': return R3h.v(T, p);
                case 'I': return R3i.v(T, p);
                case 'J': return R3j.v(T, p);
                case 'K': return R3k.v(T, p);
                case 'L': return R3l.v(T, p);
                case 'M': return R3m.v(T, p);
                case 'N': return R3n.v(T, p);
                case 'O': return R3o.v(T, p);
                case 'P': return R3p.v(T, p);
                case 'Q': return R3q.v(T, p);
                case 'R': return R3r.v(T, p);
                case 'S': return R3s.v(T, p);
                case 'T': return R3t.v(T, p);
                case 'U': return R3u.v(T, p);
                case 'V': return R3v.v(T, p);
                case 'W': return R3w.v(T, p);
                case 'X': return R3x.v(T, p);
                case 'Y': return R3y.v(T, p);
                case 'Z': return R3z.v(T, p);
                default:
                    throw std::out_of_range("Unable to match region");
            }
        }

        struct DivisionElement{
            int I;
            double n;
        };

        static DivisionElement ABdata[] = {
        {0, 0.154793642129415e4},
        {1, -0.187661219490113e3},
        {2, 0.213144632222113e2},
        {-1, -0.191887498864292e4},
        {-2, 0.918419702359447e3},
        };

        static DivisionElement CDdata[] = {
        {0, 0.585276966696349e3},
        {1, 0.278233532206915e1},
        {2, -0.127283549295878e-1},
        {3, 0.159090746562729e-3}
        };

        static DivisionElement GHdata[] = {
        {0, -0.249284240900418e5},
        {1, 0.428143584791546e4},
        {2, -0.269029173140130e3},
        {3, 0.751608051114157e1},
        {4, -0.787105249910383e-1},
        };

        static DivisionElement IJdata[] = {
        {0, 0.584814781649163e3},
        {1, -0.616179320924617},
        {2, 0.260763050899562},
        {3, -0.587071076864459e-2},
        {4, 0.515308185433082e-4},
        };

        static DivisionElement JKdata[] = {
        {0, 0.617229772068439e3},
        {1, -0.770600270141675e1},
        {2, 0.697072596851896},
        {3, -0.157391839848015e-1},
        {4, 0.137897492684194e-3},
        };

        static DivisionElement MNdata[] = {
        {0, 0.535339483742384e3},
        {1, 0.761978122720128e1},
        {2, -0.158365725441648},
        {3, 0.192871054508108e-2},
        };

        static DivisionElement OPdata[] = {
        {0, 0.969461372400213e3},
        {1, -0.332500170441278e3},
        {2, 0.642859598466067e2},
        {-1, 0.773845935768222e3},
        {-2, -0.152313732937084e4},
        };

        static DivisionElement QUdata[] = {
        {0, 0.565603648239126e3},
        {1, 0.529062258221222e1},
        {2, -0.102020639611016},
        {3, 0.122240301070145e-2},
        };

        static DivisionElement RXdata[] = {
        {0, 0.584561202520006e3},
        {1, -0.102961025163669e1},
        {2, 0.243293362700452},
        {3, -0.294905044740799e-2},
        };

        static DivisionElement UVdata[] = {
        {0, 0.528199646263062e3},
        {1, 0.890579602135307e1},
        {2, -0.222814134903755},
        {3, 0.286791682263697e-2},
        };

        static DivisionElement WXdata[] = {
        {0, 0.728052609145380e1},
        {1, 0.973505869861952e2},
        {2, 0.147370491183191e2},
        {-1, 0.329196213998375e3},
        {-2, 0.873371668682417e3},
        };

        class Region3RegionDivision{
        protected:
            std::size_t N;
            std::vector<int> I;
            std::vector<double> n;
        public:
    
            Region3RegionDivision(DivisionElement data[], std::size_t N){
                this->N = N;
                for (std::size_t i = 0; i < N; ++i){
                    n.push_back(data[i].n);
                    I.push_back(data[i].I);
                }
            }
            virtual double T_p(double p){
                const double pi = p/1e6;
                double summer = 0;
                for (std::size_t i = 0; i < N; ++i){
                    summer += n[i]*pow(pi, I[i]);
                }
                return summer*1.0;
            };
        };

        class ABline : public Region3RegionDivision{ 
            public: ABline() : Region3RegionDivision(ABdata, 5){ }; 
            virtual double T_p(double p){
                const double pi = p/1e6, ln_pi = log(pi);
                double summer = 0;
                for (std::size_t i = 0; i < N; ++i){
                    summer += n[i]*pow(ln_pi, I[i]);
                }
                return summer*1.0;
            };
        };
        class CDline : public Region3RegionDivision{ public: CDline() : Region3RegionDivision(CDdata, 4){ }; };
        class EFline { 
        public:
            double T_p(double p){ 
                const double pi = p/1e6; 
                return 3.727888004*(pi - 22.064) + 647.096; 
            }; 
        };
        class GHline : public Region3RegionDivision{ public: GHline() : Region3RegionDivision(GHdata, 5){ }; };
        class IJline : public Region3RegionDivision{ public: IJline() : Region3RegionDivision(IJdata, 5){ }; };
        class JKline : public Region3RegionDivision{ public: JKline() : Region3RegionDivision(JKdata, 5){ }; };
        class MNline : public Region3RegionDivision{ public: MNline() : Region3RegionDivision(MNdata, 4){ }; };
        class OPline : public Region3RegionDivision{ 
        public: 
            OPline() : Region3RegionDivision(OPdata, 5){ }; 
            virtual double T_p(double p){
                const double pi = p/1e6, ln_pi = log(pi);
                double summer = 0;
                for (std::size_t i = 0; i < N; ++i){
                    summer += n[i]*pow(ln_pi, I[i]);
                }
                return summer*1.0;
            };
        };
        class QUline : public Region3RegionDivision{ public: QUline() : Region3RegionDivision(QUdata, 4){ }; };
        class RXline : public Region3RegionDivision{ public: RXline() : Region3RegionDivision(RXdata, 4){ }; };
        class UVline : public Region3RegionDivision{ public: UVline() : Region3RegionDivision(UVdata, 4){ }; };
        class WXline : public Region3RegionDivision{ 
            public: WXline() : Region3RegionDivision(WXdata, 5){ }; 
            virtual double T_p(double p){
                const double pi = p/1e6, ln_pi = log(pi);
                double summer = 0;
                for (std::size_t i = 0; i < N; ++i){
                    summer += n[i]*pow(ln_pi, I[i]);
                }
                return summer*1.0;
            };
        };

        enum DividingLineEnum {LINE_AB, LINE_CD, LINE_EF, LINE_GH, LINE_IJ, LINE_JK, LINE_MN, LINE_OP, LINE_QU, LINE_RX, LINE_UV, LINE_WX};

        inline double DividingLine(DividingLineEnum region, double p){
            static ABline AB;
            static CDline CD;
            static EFline EF;
            static GHline GH;
            static IJline IJ;
            static JKline JK;
            static MNline MN;
            static OPline OP;
            static QUline QU;
            static RXline RX;
            static UVline UV;
            static WXline WX;

            switch(region){
                case LINE_AB: return AB.T_p(p);
                case LINE_CD: return CD.T_p(p);
                case LINE_EF: return EF.T_p(p);
                case LINE_GH: return GH.T_p(p);
                case LINE_IJ: return IJ.T_p(p);
                case LINE_JK: return JK.T_p(p);
                case LINE_MN: return MN.T_p(p);
                case LINE_OP: return OP.T_p(p);
                case LINE_QU: return QU.T_p(p);
                case LINE_RX: return RX.T_p(p);
                case LINE_UV: return UV.T_p(p);
                case LINE_WX: return WX.T_p(p);
                default:
                    throw std::out_of_range("Unable to match dividing line");
            }
        }
        // In the very near critical region, its messy
        inline char BackwardsRegion3SubRegionDetermination(double T, double p){

            if (p > 22.5e6){
               throw std::out_of_range("Out of range");
            }
            else if (22.11e6 < p && p <= 22.5e6){
                // Supercritical
                if (DividingLine(LINE_QU, p) < T && T <= DividingLine(LINE_UV, p)){ return 'U';}
                else if (DividingLine(LINE_UV, p) < T && T <= DividingLine(LINE_EF, p)){ return 'V';}
                else if (DividingLine(LINE_EF, p) < T && T <= DividingLine(LINE_WX, p)){ return 'W';}
                else if (DividingLine(LINE_WX, p) < T && T <= DividingLine(LINE_RX, p)){ return 'X';}
                else {return '?';}
            }
            else if (22.064e6 < p && p <= 22.11e6){
                // Supercritical
                if (DividingLine(LINE_QU, p) < T && T <= DividingLine(LINE_UV, p)){ return 'U';}
                else if (DividingLine(LINE_UV, p) < T && T <= DividingLine(LINE_EF, p)){ return 'Y';}
                else if (DividingLine(LINE_EF, p) < T && T <= DividingLine(LINE_WX, p)){ return 'Z';}
                else if (DividingLine(LINE_WX, p) < T && T <= DividingLine(LINE_RX, p)){ return 'X';}
                else {return '?';}
            }
            else if (T <= Tsat97(p)){
                if (21.93161551e6 < p && p <= 22.064e6){
                    // Sub-critical
                    if (DividingLine(LINE_QU, p) < T && T <= DividingLine(LINE_UV, p)){ return 'U';}
                    else if (DividingLine(LINE_UV, p) < T ){ return 'Y';}
                    else {return '?';}
                }
                else{
                    return 'U';
                }
            }
            else{
                if (21.90096265e6 < p && p <= 22.064e6){
                    // Sub-critical
                    if (T <= DividingLine(LINE_WX, p)){ return 'Z';}
                    else if (DividingLine(LINE_WX, p) < T && T <= DividingLine(LINE_RX, p)){ return 'X';}
                    else {return '?';}
                }
                else{
                    return 'X';
                }
            }
        }

        inline char BackwardsRegion3RegionDetermination(double T, double p){
            if (p > 100e6){
                throw std::out_of_range("pressure out of range");
            }
            else if (p > 40e6 && p <= 100e6){
                if (T <= DividingLine(LINE_AB, p)){ return 'A';}
                else {return 'B';}
            }
            else if (p > 25e6 && p <= 40e6){
                if (T <= DividingLine(LINE_CD, p)){ return 'C';}
                else if (DividingLine(LINE_CD, p) < T && T <= DividingLine(LINE_AB, p)){ return 'D';}
                else if (DividingLine(LINE_AB, p) < T && T <= DividingLine(LINE_EF, p)){ return 'E';}
                else {return 'F';}
            }
            else if (p > 23.5e6 && p <= 25e6){
                if (T <= DividingLine(LINE_CD, p)){ return 'C';}
                else if (DividingLine(LINE_CD, p) < T && T <= DividingLine(LINE_GH, p)){ return 'G';}
                else if (DividingLine(LINE_GH, p) < T && T <= DividingLine(LINE_EF, p)){ return 'H';}
                else if (DividingLine(LINE_EF, p) < T && T <= DividingLine(LINE_IJ, p)){ return 'I';}
                else if (DividingLine(LINE_IJ, p) < T && T <= DividingLine(LINE_JK, p)){ return 'J';}
                else {return 'K';}
            }
            else if (p > 23e6 && p <= 23.5e6){
                if (T <= DividingLine(LINE_CD, p)){ return 'C';}
                else if (DividingLine(LINE_CD, p) < T && T <= DividingLine(LINE_GH, p)){ return 'L';}
                else if (DividingLine(LINE_GH, p) < T && T <= DividingLine(LINE_EF, p)){ return 'H';}
                else if (DividingLine(LINE_EF, p) < T && T <= DividingLine(LINE_IJ, p)){ return 'I';}
                else if (DividingLine(LINE_IJ, p) < T && T <= DividingLine(LINE_JK, p)){ return 'J';}
                else {return 'K';}
            }
            else if (p > 22.5e6 && p <= 23e6){
                if (T <= DividingLine(LINE_CD, p)){ return 'C';}
                else if (DividingLine(LINE_CD, p) < T && T <= DividingLine(LINE_GH, p)){ return 'L';}
                else if (DividingLine(LINE_GH, p) < T && T <= DividingLine(LINE_MN, p)){ return 'M';}
                else if (DividingLine(LINE_MN, p) < T && T <= DividingLine(LINE_EF, p)){ return 'N';}
                else if (DividingLine(LINE_EF, p) < T && T <= DividingLine(LINE_OP, p)){ return 'O';}
                else if (DividingLine(LINE_OP, p) < T && T <= DividingLine(LINE_IJ, p)){ return 'P';}
                else if (DividingLine(LINE_IJ, p) < T && T <= DividingLine(LINE_JK, p)){ return 'J';}
                else {return 'K';}
            }
            else if (p > 21.04336732e6 && p <= 22.5e6){
                if (T <= DividingLine(LINE_CD, p)){ return 'C';}
                else if (DividingLine(LINE_CD, p) < T && T <= DividingLine(LINE_QU, p)){ return 'Q';}
                else if (DividingLine(LINE_RX, p) < T && T <= DividingLine(LINE_JK, p)){ return 'R';}
                else if (T > DividingLine(LINE_JK, p)){ return 'K';}
                else{ return BackwardsRegion3SubRegionDetermination(T, p);}
            }
            else if (p > 20.5e6 && p <= 21.04336732e6){
                if (T <= DividingLine(LINE_CD, p)){ return 'C';}
                else if (DividingLine(LINE_CD, p) < T && T <= Tsat97(p)){ return 'S';}
                else if (Tsat97(p) < T && T <= DividingLine(LINE_JK, p)){ return 'R';}
                else if (T > DividingLine(LINE_JK, p)){ return 'K';}
                else{ return '?';}
            }
            else if (p > 19.00881189173929e6 && p <= 20.5e6){
                if (T <= DividingLine(LINE_CD, p)){ return 'C';}
                else if (DividingLine(LINE_CD, p) < T && T <= Tsat97(p)){ return 'S';}
                else if (Tsat97(p) < T){ return 'T';}
                else{ return '?';}
            }
            else if (p > 16.529164252604481 && p <= 19.00881189173929e6){
                if (T <= Tsat97(p)){ return 'C';}
                else{ return 'T';}
            }
            else{
                return '?';
            }
        }
    } /* BackwardsRegion3 */

    /********************************************************************************/
    /**************************       Region #3       *******************************/
    /********************************************************************************/
    static RegionResidualElement Region3residdata[] = {
    {0, 0,    0.10658070028513e1},
    {0, 0,   -0.15732845290239e2},
    {0, 1,    0.20944396974307e2},
    {0, 2,   -0.76867707878716e1},
    {0, 7,    0.26185947787954e1},
    {0, 10,  -0.28080781148620e1},
    {0, 12,   0.12053369696517e1},
    {0, 23,  -0.84566812812502e-2},
    {1, 2,   -0.12654315477714e1},
    {1, 6,   -0.11524407806681e1},
    {1, 15,   0.88521043984318},
    {1, 17,  -0.64207765181607},
    {2, 0,    0.38493460186671}, 
    {2, 2,   -0.85214708824206},
    {2, 6,    0.48972281541877e1},
    {2, 7,   -0.30502617256965e1},
    {2, 22,   0.39420536879154e-1},
    {2, 26,   0.12558408424308},
    {3, 0,   -0.27999329698710},
    {3, 2,    0.13899799569460e1},
    {3, 4,   -0.20189915023570e1},
    {3, 16,  -0.82147637173963e-2},
    {3, 26,  -0.47596035734923},
    {4, 0,    0.43984074473500e-1},
    {4, 2,   -0.44476435428739},
    {4, 4,    0.90572070719733},
    {4, 26,   0.70522450087967},
    {5, 1,    0.10770512626332},
    {5, 3,   -0.32913623258954},
    {5, 26,  -0.50871062041158},
    {6, 0,   -0.22175400873096e-1},
    {6, 2,    0.94260751665092e-1},
    {6, 26,   0.16436278447961},
    {7, 2,   -0.13503372241348e-1},
    {8, 26,  -0.14834345352472e-1},
    {9, 2,    0.57922953628084e-3},
    {9, 26,   0.32308904703711e-2},
    {10, 0,   0.80964802996215e-4},
    {10, 1,  -0.16557679795037e-3},
    {11, 26, -0.44923899061815e-4},
    };

    static std::vector<RegionResidualElement> reg3rdata(Region3residdata, Region3residdata + sizeof(Region3residdata)/sizeof(RegionResidualElement));

    class Region3
    {
    protected:
        std::vector<int> Ir, Jr;
        std::vector<double> nr;
        double T_star, p_star, R;
    public:
        Region3() : T_star(1000), p_star(1e6) {
            for (std::size_t i = 0; i < reg3rdata.size(); ++i){
                nr.push_back(reg3rdata[i].n);
                Ir.push_back(reg3rdata[i].I);
                Jr.push_back(reg3rdata[i].J);
            }
            R = 461.526;
        };
        double phi(double T, double rho){
            const double rho_c = 322, T_c = 647.096, delta = rho/rho_c, tau = T_c/T;
            double summer = nr[0]*log(delta);
            for (std::size_t i = 1; i < 40; ++i){
                summer += nr[i]*pow(delta, Ir[i])*pow(tau, Jr[i]);
            }
            return summer;
        };
        double delta_dphi_ddelta(double T, double rho){
            const double rho_c = 322, T_c = 647.096, delta = rho/rho_c, tau = T_c/T;
            double summer = nr[0];
            for (std::size_t i = 1; i < 40; ++i){
                summer += nr[i]*Ir[i]*pow(delta, Ir[i])*pow(tau, Jr[i]);
            }
            return summer;
        };
        double tau_dphi_dtau(double T, double rho){
            const double rho_c = 322, T_c = 647.096, delta = rho/rho_c, tau = T_c/T;
            double summer = 0;
            for (std::size_t i = 1; i < 40; ++i){
                summer += nr[i]*Jr[i]*pow(delta, Ir[i])*pow(tau, Jr[i]);
            }
            return summer;
        };
        double delta2_d2phi_ddelta2(double T, double rho){
            const double rho_c = 322, T_c = 647.096, delta = rho/rho_c, tau = T_c/T;
            double summer = -nr[0];
            for (std::size_t i = 1; i < 40; ++i){
                summer += nr[i]*Ir[i]*(Ir[i]-1)*pow(delta, Ir[i])*pow(tau, Jr[i]);
            }
            return summer;
        };
        double tau2_d2phi_dtau2(double T, double rho){
            const double rho_c = 322, T_c = 647.096, delta = rho/rho_c, tau = T_c/T;
            double summer = 0;
            for (std::size_t i = 1; i < 40; ++i){
                summer += nr[i]*Jr[i]*(Jr[i]-1)*pow(delta, Ir[i])*pow(tau, Jr[i]);
            }
            return summer;
        };
        double deltatau_d2phi_ddelta_dtau(double T, double rho){
            const double rho_c = 322, T_c = 647.096, delta = rho/rho_c, tau = T_c/T;
            double summer = 0;
            for (std::size_t i = 1; i < 40; ++i){
                summer += nr[i]*Jr[i]*Ir[i]*pow(delta, Ir[i])*pow(tau, Jr[i]);
            }
            return summer;
        };
        double p(double T, double rho){
            return rho*R*T*delta_dphi_ddelta(T, rho);
        };
        double umass(double T, double rho){
            return R*T*tau_dphi_dtau(T, rho);
        };
        double smass(double T, double rho){
            return R*(tau_dphi_dtau(T, rho) - phi(T, rho));
        };
        double hmass(double T, double rho){
            return R*T*(tau_dphi_dtau(T, rho) + delta_dphi_ddelta(T, rho));
        };
        double cpmass(double T, double rho){
            return R*(-tau2_d2phi_dtau2(T, rho) + pow(delta_dphi_ddelta(T, rho) - deltatau_d2phi_ddelta_dtau(T, rho), 2)/(2*delta_dphi_ddelta(T, rho) + delta2_d2phi_ddelta2(T, rho)));
        };
        double cvmass(double T, double rho){
            return R*(-tau2_d2phi_dtau2(T, rho));
        };
        double speed_sound(double T, double rho){
            double RHS = 2*delta_dphi_ddelta(T, rho) + delta2_d2phi_ddelta2(T, rho)-pow(delta_dphi_ddelta(T,rho)-deltatau_d2phi_ddelta_dtau(T,rho),2)/tau2_d2phi_dtau2(T,rho);
            return sqrt(R*T*RHS);
        };
        double output(I97parameters key, double T, double p){
            char region = Region3Backwards::BackwardsRegion3RegionDetermination(T, p);
            double rho = 1/Region3Backwards::Region3_v_TP(region, T, p);
            switch(key){
            case IF97_DMASS: return rho;
            case IF97_HMASS: return hmass(T, rho);
            case IF97_SMASS: return smass(T, rho);
            case IF97_UMASS: return umass(T, rho);
            case IF97_CPMASS: return cpmass(T, rho);
            case IF97_CVMASS: return cvmass(T, rho);
            case IF97_W: return speed_sound(T, rho);
            default:
                throw std::out_of_range("bad key to output");
            }
        }
    };

    /********************************************************************************/
    /**************************       Region #4       *******************************/
    /********************************************************************************/
    struct SaturationElement{
        int i;
        double n;
    };
    static SaturationElement sat[] = {
        {1,  0.11670521452767e4},
        {2, -0.72421316703206e6},
        {3, -0.17073846940092e2},
        {4,  0.12020824702470e5},
        {5, -0.32325550322333e7},
        {6,  0.14915108613530e2},
        {7, -0.48232657361591e4},
        {8,  0.40511340542057e6},
        {9, -0.23855557567849},
        {10, 0.65017534844798e3},
    };
    static std::vector<SaturationElement> reg4data(sat, sat + sizeof(sat)/sizeof(SaturationElement));
    /// This "region" is the saturation curve
    class Region4
    {
    public:
        std::vector<double> n;
        double p_star, T_star;

        Region4() : p_star(1e6), T_star(1.0){
            n.resize(1); n[0] = 0;
            for (std::size_t i = 0; i < reg4data.size(); ++i){
                n.push_back(reg4data[i].n);
            }
        };
        double p_T(double T){
            double theta = T/T_star+n[9]/(T/T_star-n[10]);
            double A = theta*theta + n[1] * theta + n[2];
            double B = n[3]*theta*theta + n[4]*theta + n[5];
            double C = n[6]*theta*theta + n[7]*theta + n[8];
            return p_star*pow(2*C/(-B+sqrt(B*B-4*A*C)), 4);
        };
        double T_p(double p){
            double beta = pow(p/p_star, 0.25);
            double E = beta*beta + n[3]*beta + n[6];
            double F = n[1]*beta*beta + n[4]*beta + n[7];
            double G = n[2]*beta*beta + n[5]*beta + n[8];
            double D = 2*G/(-F-sqrt(F*F-4*E*G));
            return T_star*0.5*(n[10] + D - sqrt(pow(n[10]+D, 2) - 4*(n[9] + n[10]*D)));
        };
    };

    /********************************************************************************/
    /**************************       Region #5       *******************************/
    /********************************************************************************/
    static RegionResidualElement Region5residdata[] = {
        {1, 1,  0.15736404855259e-2},
        {1, 2,  0.90153761673944e-3},
        {1, 3, -0.50270077677648e-2},
        {2, 3,  0.22440037409485e-5},
        {2, 9, -0.41163275453471e-5},
        {3, 7,  0.37919454822955e-7}
    };
    static RegionIdealElement Region5idealdata[] = {
        { 0, -0.13179983674201e2},
        { 1,  0.68540841634434e1},
        {-3, -0.24805148933466e-1},
        {-2,  0.36901534980333},
        {-1, -0.31161318213925e1},
        { 2, -0.32961626538917}
    };
    static std::vector<RegionResidualElement> reg5rdata(Region5residdata, Region5residdata + sizeof(Region5residdata)/sizeof(RegionResidualElement));
    static std::vector<RegionIdealElement> reg50data(Region5idealdata, Region5idealdata + sizeof(Region5idealdata)/sizeof(RegionIdealElement));

    class Region5 : public BaseRegion
    {
    public:
        Region5() : BaseRegion(reg5rdata, reg50data)  {
            T_star = 1000; p_star = 1e6; 
        };
        double TAUrterm(double T){
            return T_star/T;
        }
        double PIrterm(double p){
            return p/p_star;
        }
        double TAU0term(double T){return T_star/T;}
    };

    enum IF97REGIONS {REGION_1, REGION_2, REGION_3, REGION_4, REGION_5};

    inline IF97REGIONS RegionDetermination_TP(double T, double p)
    {
        static Region4 R4;
        if (T > 2273.15){
            throw std::out_of_range("Out of range");
        }
        else if (T > 1073.15 && T < 2273.15){
            if (p < 50e6){
                return REGION_5;
            }
            else{
                throw std::out_of_range("Out of range");
            }
        }
        else if (T > 623.15 && T < 1073.15){
            if (p > 100e6){
                throw std::out_of_range("Out of range");
            }
            else if (p < 16.5292e6){ // Check this one first to avoid the call to 2-3 boundary curve (a little bit faster)
                return REGION_2;
            }
            else if (p > Region23_T(T)){
                return REGION_3;
            }
            else{
                return REGION_2;
            }
        }
        else if (T > 273.15 && T < 623.15){
            if (p > 100e6){
                throw std::out_of_range("Out of range");
            }
            else if(p > R4.p_T(T)){
                return REGION_1;
            }
            else{
                return REGION_2;
            }
        }
        else{
            throw std::out_of_range("Out of range");
        }
    }

    inline double RegionOutput(IF97REGIONS region, I97parameters outkey, double T, double p){
        static Region1 R1;
        static Region2 R2;
        static Region3 R3;
        static Region4 R4;
        static Region5 R5;
    
        switch (region){
            case REGION_1: return R1.output(outkey, T, p);
            case REGION_2: return R2.output(outkey, T, p);
            case REGION_3: return R3.output(outkey, T, p);
            case REGION_4: throw std::out_of_range("Cannot use Region 4 with T and p as inputs");
            case REGION_5: return R5.output(outkey, T, p);
        }
        throw std::out_of_range("Unable to match region");
    }
    
    // ******************************************************************************** //
    //                                     API
    // ******************************************************************************** //

    /// Get the mass density [kg/m^3] as a function of T [K] and p [Pa]
    inline double rhomass_Tp(double T, double p){ return RegionOutput(RegionDetermination_TP(T, p), IF97_DMASS, T, p); };
    /// Get the mass enthalpy [J/kg] as a function of T [K] and p [Pa]
    inline double hmass_Tp(double T, double p){ return RegionOutput(RegionDetermination_TP(T, p), IF97_HMASS, T, p); };
    /// Get the mass entropy [J/kg/K] as a function of T [K] and p [Pa]
    inline double smass_Tp(double T, double p){ return RegionOutput(RegionDetermination_TP(T, p), IF97_SMASS, T, p); };
    /// Get the mass internal energy [J/kg] as a function of T [K] and p [Pa]
    inline double umass_Tp(double T, double p){ return RegionOutput(RegionDetermination_TP(T, p), IF97_UMASS, T, p); };
    /// Get the mass constant-pressure specific heat [J/kg/K] as a function of T [K] and p [Pa]
    inline double cpmass_Tp(double T, double p){ return RegionOutput(RegionDetermination_TP(T, p), IF97_CPMASS, T, p); };
    /// Get the mass constant-volume specific heat [J/kg/K] as a function of T [K] and p [Pa]
    inline double cvmass_Tp(double T, double p){ return RegionOutput(RegionDetermination_TP(T, p), IF97_CVMASS, T, p); };
    /// Get the speed of sound [m/s] as a function of T [K] and p [Pa]
    inline double speed_sound_Tp(double T, double p){ return RegionOutput(RegionDetermination_TP(T, p), IF97_W, T, p); };
    /// Get the saturation temperature [K] as a function of p [Pa]
    inline double Tsat97(double p){
        static Region4 R4;
        return R4.T_p(p);
    }
    /// Get the saturation pressure [Pa] as a function of T [K]
    inline double psat97(double T){
        static Region4 R4;
        return R4.p_T(T);
    }
}; /* namespace IF97 */

#if defined(ENABLE_CATCH)

struct Region3BackwardsData{
    char region; double T, p, v;
};

static Region3BackwardsData _Table5[] = {
    // Table 5 data 
    {'A', 630, 50e6, 1.470853100e-3},
    {'A', 670, 80e6, 1.503831359e-3},
    {'B', 710, 50e6, 2.204728587e-3},
    {'B', 750, 80e6, 1.973692940e-3},
    {'C', 630, 20e6, 1.761696406e-3},
    {'C', 650, 30e6, 1.819560617e-3},
    {'D', 656, 26e6, 2.245587720e-3},
    {'D', 670, 30e6, 2.506897702e-3},
    {'E', 661, 26e6, 2.970225962e-3},
    {'E', 675, 30e6, 3.004627086e-3},
    {'F', 671, 26e6, 5.019029401e-3},
    {'F', 690, 30e6, 4.656470142e-3},
    {'G', 649, 23.6e6, 2.163198378e-3},
    {'G', 650, 24e6, 2.166044161e-3},
    {'H', 652, 23.6e6, 2.651081407e-3},
    {'H', 654, 24e6, 2.967802335e-3},
    {'I', 653, 23.6e6, 3.273916816e-3},
    {'I', 655, 24e6, 3.550329864e-3},
    {'J', 655, 23.5e6, 4.545001142e-3},
    {'J', 660, 24e6, 5.100267704e-3},
    {'K', 660, 23e6, 6.109525997e-3},
    {'K', 670, 24e6, 6.427325645e-3},
    {'L', 646, 22.6e6, 2.117860851e-3},
    {'L', 646, 23e6, 2.062374674e-3},
    {'M', 648.6, 22.6e6, 2.533063780e-3},
    {'M', 649.3, 22.8e6, 2.572971781e-3},
    {'N', 649.0, 22.6e6, 2.923432711e-3},
    {'N', 649.7, 22.8e6, 2.913311494e-3},
    {'O', 649.1, 22.6e6, 3.131208996e-3},
    {'O', 649.9, 22.8e6, 3.221160278e-3},
    {'P', 649.4, 22.6e6, 3.715596186e-3},
    {'P', 650.2, 22.8e6, 3.664754790e-3},
    {'Q', 640, 21.1e6, 1.970999272e-3},
    {'Q', 643, 21.8e6, 2.043919161e-3},
    {'R', 644, 21.1e6, 5.251009921e-3},
    {'R', 648, 21.8e6, 5.256844741e-3},
    {'S', 635, 19.1e6, 1.932829079e-3},
    {'S', 638, 20e6, 1.985387227e-3},
    {'T', 626, 17e6, 8.483262001e-3},
    {'T', 640, 20e6, 6.227528101e-3},
    // Table 11
    {'U', 644.6, 21.5e6, 2.268366647e-3},
    {'U', 646.1, 22e6, 2.296350553e-3},
    {'V', 648.6, 22.5e6, 2.832373260e-3},
    {'V', 647.9, 22.3e6, 2.811424405e-3},
    {'W', 647.5, 22.15e6, 3.694032281e-3},
    {'W', 648.1, 22.3e6, 3.622226305e-3},
    {'X', 648, 22.11e6, 4.528072649e-3},
    {'X', 649, 22.3e6, 4.556905799e-3},
    {'Y', 646.84, 22e6, 2.698354719e-3},
    {'Y', 647.05, 22.064e6, 2.717655648e-3},
    {'Z', 646.89, 22e6, 3.798732962e-3},
    {'Z', 647.15, 22.064e6, 3.701940010e-3},
};


inline void print_IF97_Table5()
{
    static std::vector<Region3BackwardsData> Table5(_Table5, _Table5 + sizeof(_Table5)/sizeof(Region3BackwardsData));
    for (std::size_t i = 0; i < Table5.size(); ++i){
        double v = IF97::Region3Backwards::Region3_v_TP(Table5[i].region, Table5[i].T, Table5[i].p);
        if (std::abs(v - Table5[i].v) > 1e-12){
            std::cout << Table5[i].region << " " << Table5[i].T << " " << Table5[i].p << " " << v << " " << Table5[i].v << std::endl;
        }
        char region = IF97::Region3Backwards::BackwardsRegion3RegionDetermination(Table5[i].T, Table5[i].p);
        if (region != Table5[i].region)
        {
            IF97::Region3Backwards::BackwardsRegion3RegionDetermination(Table5[i].T, Table5[i].p);
            std::cout << std::setprecision(10) << Table5[i].region << " " << region << " " << Table5[i].T << " " << Table5[i].p << " " << v << " " << Table5[i].v << std::endl;
        }
    }
};

struct Table3Data{
    IF97::Region3Backwards::DividingLineEnum region; 
    double p, T;
};

static Table3Data _Table3[] = {
    
    // Table 3
    {IF97::Region3Backwards::LINE_AB, 40e6, 6.930341408e2},
    {IF97::Region3Backwards::LINE_CD, 25e6, 6.493659208e2},
    {IF97::Region3Backwards::LINE_EF, 40e6, 7.139593992e2},
    {IF97::Region3Backwards::LINE_GH, 23e6, 6.498873759e2},
    {IF97::Region3Backwards::LINE_IJ, 23e6, 6.515778091e2},
    {IF97::Region3Backwards::LINE_JK, 23e6, 6.558338344e2},
    {IF97::Region3Backwards::LINE_MN, 22.8e6, 6.496054133e2},
    {IF97::Region3Backwards::LINE_OP, 22.8e6, 6.500106943e2},
    {IF97::Region3Backwards::LINE_QU, 22e6, 6.456355027e2},
    {IF97::Region3Backwards::LINE_RX, 22e6, 6.482622754e2},
    // Table 11
    {IF97::Region3Backwards::LINE_UV, 22.3e6, 6.477996121e2},
    {IF97::Region3Backwards::LINE_WX, 22.3e6, 6.482049480e2},
};
static std::vector<Table3Data> Table3(_Table3, _Table3 + sizeof(_Table3)/sizeof(Table3Data));

inline void print_boundary_line_Table3()
{
    for (std::size_t i = 0; i < Table3.size(); ++i){
        double T = IF97::Region3Backwards::DividingLine(Table3[i].region, Table3[i].p);
        if (std::abs(T - Table3[i].T) > 1e-7){
            printf("p: %g errT: %g\n", Table3[i].p, T - Table3[i].T);
        }
    }
};

#endif

#endif
